local ScenarioUtils = import('/lua/sim/ScenarioUtilities.lua');
local ScenarioFramework = import('/lua/ScenarioFramework.lua');
local Utilities = import('/lua/utilities.lua');
local AIUtils = import('/lua/AI/aiutilities.lua')

##Original script by FunkOff(Map Survial Run v3) w/ some v3 code; Modified by DGoReck(Map: Survival Escape)
##Version 2.3    8.02.2009

local TargetWaypoint = 0  #Do not touch plx
local TotalWaypoints = nil  #Map maker needs not fuck with this any longer
local TotalBossBattles = nil #Map maker needs not input this value either
local TotalExpansionPoints = nil  #The number of playable areas - 1.  Don't bother adjusting this.
local CurrentArea = 1

#local TotalWaypoints = 11;  #Map maker needs to input the total number of waypoints excluding WAYPOINT_0
#local TotalBossBattles = 5; #Map maker needs to specify the total number of boss battles supported, put 0 for none
#local TotalExpansionPoints = 2;  ##The number of playable areas - 1.  This lets the script know when to check for map expansion.  If no map expansion points, put 0.


local oldOnPopulate = OnPopulate

   function OnPopulate()
        #oldOnPopulate()
        ScenarioUtils.InitializeArmies()
	Survival_InitGame();
 	ScenarioFramework.SetPlayableArea('AREA_1', false);
   end

   local oldOnStart = OnStart
   function OnStart(self)
      oldOnStart(self)
   end
 

-- econ adjust based on who is playing
-- taken from original survival/Jotto
--------------------------------------------------------------------------
function ScenarioUtils.CreateResources()
	local Markers = ScenarioUtils.GetMarkers();

	for i, tblData in pairs(Markers) do -- loop marker list
		local SpawnThisResource = false; -- default to no
		if (tblData.resource and not tblData.SpawnWithArmy) then -- if this is a regular resource
			SpawnThisResource = true;
		elseif (tblData.resource and tblData.SpawnWithArmy) then -- if this is an army-specific resource
			if (tblData.SpawnWithArmy == "ARMY_0") then
				SpawnThisResource = true;
			else
				for x, army in ListArmies() do -- loop through army list

					if (tblData.SpawnWithArmy == army) then -- if this army is present
						SpawnThisResource = true; -- spawn this resource
						break;
					end
				end
			end
		end

		if (SpawnThisResource) then -- if we can spawn the resource do it
			local bp, albedo, sx, sz, lod;
			if (tblData.type == "Mass") then
				albedo = "/env/common/splats/mass_marker.dds";
				bp = "/env/common/props/massDeposit01_prop.bp";
				sx = 2;
				sz = 2;
				lod = 100;
			else
				albedo = "/env/common/splats/hydrocarbon_marker.dds";
				bp = "/env/common/props/hydrocarbonDeposit01_prop.bp";
				sx = 6;
				sz = 6;
				lod = 200;
			end
			-- create the resource
			CreateResourceDeposit(tblData.type,	tblData.position[1], tblData.position[2], tblData.position[3], tblData.size);

			-- create the resource graphic on the map
			CreatePropHPR(bp, tblData.position[1], tblData.position[2], tblData.position[3], Random(0,360), 0, 0);

			-- create the resource icon on the map
			CreateSplat(
				tblData.position,           # Position
				0,                          # Heading (rotation)
				albedo,                     # Texture name for albedo
				sx, sz,                     # SizeX/Z
				lod,                        # LOD
				0,                          # Duration (0 == does not expire)
				-1,                         # army (-1 == not owned by any single army)
				0							# ???
			);
		end
	end
end  
  
  
-- broadcast a text message to players
----------------------------------------------------------------------------
BroadcastMSG = function(MSG, Fade, TextColor)
	PrintText(MSG, 20, TextColor, Fade, 'center');
end


-- initializes the game settings
--------------------------------------------------------------------------
Survival_InitGame = function()
	LOG("-------Finding important Markers------")
	local WaypointMarkers = {}
	local BossMarkers = {}
	local BossSpawnMarkers = {}
	local MapExpansionMarkers = {}
    local Markers = ScenarioUtils.GetMarkers()
	for i,name in Markers do
		if string.sub(i,1,9) == "WAYPOINT_" then
			table.insert(WaypointMarkers, i)
		elseif string.sub(i,1,12) == "BOSS_BATTLE_" then
			table.insert(BossMarkers, i)
		elseif string.sub(i,1,14) == "MAP_EXPANSION_"  then
			table.insert(MapExpansionMarkers, i)
		elseif string.sub(i,1,10) == "BOSSSPAWN_" then
			table.insert(BossSpawnMarkers, i)
		end
	end
	TotalWaypoints = table.getn(WaypointMarkers)
	TotalWaypoints = TotalWaypoints - 1 #adjusting for Waypoint 0
	LOG("There are a total of " .. TotalWaypoints - 1 .. " waypoints")
	TotalBossBattles = table.getn(BossMarkers)
	LOG("There are a total of " .. TotalBossBattles .. " Boss Battles")
	TotalMapExpansions = table.getn(MapExpansionMarkers)
	LOG("There are a total of " .. TotalMapExpansions .. " map expansions")
	LOG("----- Survival MOD: Configuring match settings...")
	-- check game configuration	
	
	SetArmyColor('ARMY_SURVIVAL_ENEMY',150,0,25) 
	SetArmyColor('ARMY_SURVIVAL_ALLY',0,150,50) 
	
	local Difficulty = ScenarioInfo.Options.Difficulty;

	if Difficulty == nil then
		Difficulty = 4;
	end
	LOG("Difficulty is " .. Difficulty);
		
 	ScenarioInfo.Options.Victory = 'sandbox'; -- force sandbox in order to implement our own rules

	local Armies = ListArmies();
	local ArmiesPresent = {}
	Survival_PlayerCount = table.getn(Armies) - 2; -- save player count, subtracting the 2 AI "players"

	-- loop through armies
	for i, Army in ListArmies() do
		if (Army == "ARMY_1") then
			ArmiesPresent[1] = true
		elseif (Army == "ARMY_2") then
			ArmiesPresent[2] = true
		elseif (Army == "ARMY_3") then
			ArmiesPresent[3] = true
		elseif	(Army == "ARMY_4") then
			ArmiesPresent[4] = true
		end
		-- check if it's a human army
		if (Army == "ARMY_1" or Army == "ARMY_2" or Army == "ARMY_3" or Army == "ARMY_4" or Army == "ARMY_5" or Army == "ARMY_6" or Army == "ARMY_7" or Army == "ARMY_8") then 

			ScenarioFramework.AddRestriction(Army, categories.WALL); -- don't allow them to build walls
			#ScenarioFramework.AddRestriction(Army, categories.AIR); -- don't allow them to build air stuff

			-- loop through other armies to ally with other human armies
			for x, ArmyX in ListArmies() do
				-- if human army
				if (ArmyX == "ARMY_1" or ArmyX == "ARMY_2" or ArmyX == "ARMY_3" or ArmyX == "ARMY_4" or ArmyX == "ARMY_5" or ArmyX == "ARMY_6" or ArmyX == "ARMY_7" or ArmyX == "ARMY_8") then 
					SetAlliance(Army, ArmyX, 'Ally'); 
				end
			end			

			SetAlliance(Army, "ARMY_SURVIVAL_ALLY", 'Ally'); -- friendly AI team
			SetAlliance(Army, "ARMY_SURVIVAL_ENEMY", 'Enemy');  -- enemy AI team

			SetAlliedVictory(Army, true); -- can win together of course :)
		end
	end

	SetAlliance("ARMY_SURVIVAL_ALLY", "ARMY_SURVIVAL_ENEMY", 'Enemy'); -- the friendly and enemy AI teams should be enemies
	SetArmyUnitCap("ARMY_SURVIVAL_ENEMY", 10000)
	SetArmyUnitCap("ARMY_SURVIVAL_ALLY", 10000)
	
		--Log("Current starting area is " .. CurrentArea)
		--SpawnDefenses(CurrentArea) --need to implement
		
		ForkThread(Survival_SpawnDef)
		LOG("Spawned Defensive Target");
		--Survival_SpawnDef();
		
		ForkThread(SpawnEnemies, Difficulty);
		Survival_SpawnEnemyRandoms();  -- work in progress.
		
		ForkThread(EmergencyShield, Difficulty); 
		ForkThread(Immunize_StartUnits)

		if TotalBossBattles > 0 then
			ForkThread(BossBattles, BossSpawnMarkers);
		end
		LOG("Checking Boss Battles status");
		
		ForkThread(ExpansionCheck);
		LOG("Checking to see if map should be expanded");
		
		LOG("Spawning Ally Units in-lui of playes");
		
		local ArmyTechUnit = {
		[1] = {
			[1] = 'uel0103',
			[2] = 'uel0111',
			[3] = 'uel0304',
			[0] = 'ueb0304',
			},
		[2] = {
			[1] = 'uel0201',
			[2] = 'uel0202',
			[3] = 'uel0303',
			[0] = 'ueb0304',
			},
		[3] = {
			[1] = 'uaa0102',
			[2] = 'xaa0202',
			[3] = 'uaa0303',
			[0] = 'uab0304',
			},
		[4] = {
			[1] = 'xra0105',
			[2] = 'ura0203',
			[3] = 'xra0305',
			[0] = 'urb0304',
			},		
		}
		
		for n=1, 4 do
			if not ArmiesPresent[n] == true then
					ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ALLY','GROUP_' .. repr(n))
					--ForkThread(ArmiesMissingSupport,n,ArmyTechUnit[n], Difficulty)  --Need to implement
			end
		end
		
end



-- Function to prevent friendly fire of ALLY civ AI to Human armies ??
-------------------------------------------------------------------------
newOnStartBuild = function(self, unitBeingBuilt, order)
   self:OldOnStartBuild(unitBeingBuilt, order )
   
   ImmunizeUnit(unitBeingBuilt)
end
--Replacement Collision Check function
newOnCollisionCheck = function(self, other, firingWeapon)
   if other.friendly == 1 then
      return false
   else
      return self.OldOnCollisionCheck(self, other, firingWeapon)   
   end
end
--Replacement function for Projectile Creation
newCreateProjectileForWeapon = function(self, muzzle)
   local proj = self.OldCreateProjectileForWeapon(self, muzzle)
   proj.friendly = 1
   return proj
end
--Replacement function for Child Projectile Creation
newCreateChildProjectile = function(self, bp)
   local proj = self.OldCreateChildProjectile(bp)
   proj.friendly = 1
   return proj
end
--Replacement Collision Check function used by Beam Weapons
newOnCollisionCheckWeapon = function(self, firingWeapon)
   if firingWeapon.friendly == 1 then
      return false
   else
      return self.OldOnCollisionCheckWeapon(self, firingWeapon)   
   end
end
--Mark weapons as friendly
neutralizeWeapons = function(unit)
   for i = 1, unit:GetWeaponCount() do
      local weapon = unit:GetWeapon(i)
      local bp = unit:GetBlueprint().Weapon[i]
      if bp.ProjectileId != "" then
         weapon.OldCreateProjectileForWeapon = weapon.CreateProjectileForWeapon
         weapon.CreateProjectileForWeapon = newCreateProjectileForWeapon
         weapon.OldCreateChildProjectile = weapon.CreateChildProjectile
         weapon.CreateChildProjectile = newCreateChildProjectile
      else
         weapon.friendly = 1
      end
   end
end


Waypoint_Tracker = function()
	--local TargetWaypoint = GetClosestWaypoint()  --Should be 0 at the start of the game
	local OldDistance = nil
	local PlatoonList = {}
	table.insert(PlatoonList, Survival_DefUnit) -- add unit to a platoon
	local aiBrain = GetArmyBrain("ARMY_SURVIVAL_ALLY") --GetArmyBrain(ArmyID)	--"ARMY_SURVIVAL_ENEMY")
	local aiPlatoon = aiBrain:MakePlatoon('','')
	aiBrain:AssignUnitsToPlatoon(aiPlatoon, PlatoonList, 'Attack', 'None') -- platoon, unit list, "mission" and 

	BroadcastMSG("Spider Pig is on the move! Defend it. HQ out.",5)

	while TargetWaypoint < TotalWaypoints + 1 do
		local Distance = VDist3(Survival_DefUnit:GetPosition(),ScenarioUtils.MarkerToPosition("WAYPOINT_" .. TargetWaypoint))	

		if Distance < 10 then
			TargetWaypoint = TargetWaypoint + 1
			aiPlatoon:MoveToLocation(ScenarioUtils.MarkerToPosition("WAYPOINT_" .. (TargetWaypoint)), false)
			LOG("Truck has qued a move order to waypoint number " .. TargetWaypoint)			
		elseif Distance == OldDistance then
			--TargetWaypoint = GetClosestWaypoint() + 1
			aiPlatoon:MoveToLocation(ScenarioUtils.MarkerToPosition("WAYPOINT_" .. TargetWaypoint), false)	
			LOG("Truck was stuck. Re-queing a move order to waypoint number " .. TargetWaypoint)
		end
		LOG("Target waypoint is waypoint " .. TargetWaypoint .. " at distance " .. Distance)
		OldDistance = Distance
		WaitSeconds(10)	
	end	
end

-- function to check if HQ reached final waypoint
---------------------------------------------------------------------
function VictoryCheck()

	while not Survival_DefUnit:IsDead() do
		local Distance = VDist3(Survival_DefUnit:GetPosition(),ScenarioUtils.MarkerToPosition("GOAL"));
			if (Distance <= 10) then
			BroadcastMSG("Spider Pig has reached it's destination.  We will be able to build Dr Whos weapons new weapon soo.  HQ out.", 4);
			Survival_GameState = 2;
				for i, army in ListArmies() do
					if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4" or army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then
					GetArmyBrain(army):OnVictory();
					end
				end
			GetArmyBrain("ARMY_SURVIVAL_ENEMY"):OnDefeat();
			end
	WaitSeconds(10)
	end
end

--prevents units from hitting friendly units
--------------------------------------------
Immunize_StartUnits = function()
   WaitSeconds(5)
   for i, army in ListArmies() do
      if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4" or army == "ARMY_SURVIVAL_ALLY") then
         local unitList = GetArmyBrain(army):GetListOfUnits( categories.ALLUNITS , false, false)
         for j, unit in unitList do
            ImmunizeUnit(unit)
         end
      end   
   end
end

## Makes a unit immune to friendly fire and neutralizes its weapons to not harm allies
ImmunizeUnit = function(unit)
   if unit.immune == nil then
      unit.immune = 1
      -- Makes friendly unit's projectiles pass through this unit
      unit.OldOnCollisionCheck = unit.OnCollisionCheck
      unit.OnCollisionCheck = newOnCollisionCheck
      unit.OldOnCollisionCheckWeapon = unit.OnCollisionCheckWeapon
      unit.OnCollisionCheckWeapon = newOnCollisionCheckWeapon
      
      -- Marks this unit's weapons as friendly
      neutralizeWeapons(unit)
      
      -- Modify unit's build function to pass immunity on to other units
      if unit.OnStartBuild and unit.OldOnStartBuild == nil then
         unit.OldOnStartBuild = unit.OnStartBuild
         unit.OnStartBuild = newOnStartBuild
      end
   end
end



-- function to spawn enemy units
---------------------------------------------------------------------
function SpawnEnemies(Difficulty)

	local StartTime = ScenarioInfo.Options.StartTime;

	if StartTime == nil then
		StartTime = 0;
	end

	BroadcastMSG("Spider Pig Transport is arriving in " .. (StartTime + 5) .. " seconds.", 5);

	Survival_DefUnit:SetSpeedMult(0.01);
	WaitSeconds(StartTime + 5);
	Survival_DefUnit:SetSpeedMult(0.5);

	BroadcastMSG("Spider Pig is on the move.  Defend it.  HQ out.", 5);
	WaitSeconds(15);

	-- Not implemented yet
--	local ReinforcementStrength = ScenarioInfo.Options.Reinforcements
--	if ReinforcementStrength == nil then
--		ReinforcementStrength = 1
--	end
--	if ReinforcementStrength > 0 then
--		ForkThread(Reinforcements, ReinforcementStrength)
--	end
	
	local SpawnMultiplier = nil;
	local WaveDelay = nil;

	if Difficulty == 1 then
		WaveDelay = 50;
		SpawnMultiplier = .5;
	elseif Difficulty == 2 then
		WaveDelay = 40;
		SpawnMultiplier = .8;
	elseif Difficulty == 3 then
		WaveDelay = 30;
		SpawnMultiplier = 1;
	elseif Difficulty == 4 then
		WaveDelay = 20;
		SpawnMultiplier = 1;
	elseif Difficulty == 5 then
		WaveDelay = 15;
		SpawnMultiplier = 1.5;
	elseif Difficulty == 6 then
		WaveDelay = 12;
		SpawnMultiplier = 1.5;
	elseif Difficulty == 7 then
		WaveDelay = 10;
		SpawnMultiplier = 2;
	elseif Difficulty == 8 then
		WaveDelay = 8;
		SpawnMultiplier = 2.5;
	else
		WaveDelay = 5;
		SpawnMultiplier = 3;
	end

	BroadcastMSG("The Cyprst Milita are attacking us.  Destroy them.  HQ Out.", 5)


	local WaypointNumber = 0;
	local StartingGates = {};

	-- ground forces starting gates
	for n=0, TotalWaypoints do  -- was 1
   	  local units =  AIUtils.GetOwnUnitsAroundPoint(GetArmyBrain("ARMY_SURVIVAL_ENEMY"), categories.urb0304, ScenarioUtils.MarkerToPosition("GROUNDSPAWN_" .. repr(n)), 30 )
   	   StartingGates[ n ] = 0;
   	   for _,unit in units do
       	         StartingGates[ n ] = StartingGates[ n ] + 1;
  	   end
   	   LOG("Total gates around groundspawn "..repr(n).." is "..repr(StartingGates[ n ]));
	end
	
	-- naval forces starting gates
	local NavalStartingGates = {};

	for p=0, TotalWaypoints do
    	  local pUnits =  AIUtils.GetOwnUnitsAroundPoint(GetArmyBrain("ARMY_SURVIVAL_ENEMY"), categories.urb0303, ScenarioUtils.MarkerToPosition("NAVALSPAWN_" .. repr(p)), 30 )
   	   NavalStartingGates[ p ] = 0;
   	   for _,unit in pUnits do
       	         NavalStartingGates[ p ] = NavalStartingGates[ p ] + 1;
  	   end
   	   LOG("Total gates around navalspawn "..repr(p).." is "..repr(NavalStartingGates[ p ]));
	end



	while not Survival_DefUnit:IsDead() do

		--local Distance = VDist3(Survival_DefUnit:GetPosition(),ScenarioUtils.MarkerToPosition("GOAL"));
		--local ClosestWaypoint = GetClosestWaypoint();
		--LOG("Closest waypoint is waypoint " .. ClosestWaypoint);
		SpawnGroundWave(TargetWaypoint, SpawnMultiplier, StartingGates[ TargetWaypoint ])
		SpawnAirWave(TargetWaypoint, SpawnMultiplier)
		SpawnNavalWave(TargetWaypoint, SpawnMultiplier, NavalStartingGates[ TargetWaypoint ])

		if TargetWaypoint > 1 then
			SpawnGroundWave((TargetWaypoint - 1), SpawnMultiplier, StartingGates[ TargetWaypoint - 1])
			SpawnAirWave((TargetWaypoint - 1), SpawnMultiplier)
			SpawnNavalWave((TargetWaypoint - 1), SpawnMultiplier, NavalStartingGates[ TargetWaypoint ])
		end

		if TargetWaypoint > 2 then
			SpawnGroundWave((TargetWaypoint - 2), SpawnMultiplier, StartingGates[ TargetWaypoint - 2 ])
			SpawnAirWave((TargetWaypoint - 2), SpawnMultiplier)
			SpawnNavalWave((TargetWaypoint - 2), SpawnMultiplier, NavalStartingGates[ TargetWaypoint ])
		end

		if TargetWaypoint > 3 then
			SpawnGroundWave((TargetWaypoint- 3), SpawnMultiplier, StartingGates[ TargetWaypoint - 3 ])
			SpawnAirWave((TargetWaypoint - 3), SpawnMultiplier)
			SpawnNavalWave((TargetWaypoint - 3), SpawnMultiplier, NavalStartingGates[ TargetWaypoint ])
		end

		WaitSeconds(WaveDelay)

	end


end


-- Determine which waypoint the defense object is closest to so that way we can spawn enemies from corresponding spawn points
----------------------------
function GetClosestWaypoint(POS)
	
	if POS == nil then
		POS = Survival_DefUnit:GetPosition()
	end

	local WaypointCounter = 0;
	local ShortestDistance = 20000;
	local ShortestDistanceWaypoint = 100;
	local WaypointDistance = 1000;

	while (WaypointCounter < (TotalWaypoints + 1)) do
		WaypointDistance = VDist3(POS,ScenarioUtils.MarkerToPosition("WAYPOINT_" .. WaypointCounter))
		if WaypointDistance < ShortestDistance then
			ShortestDistance = WaypointDistance;
			ShortestDistanceWaypoint = WaypointCounter;
		end
		WaypointCounter = WaypointCounter + 1;
	end
	return ShortestDistanceWaypoint;
end


-- Ground spawn
---------------------------------------------------------------------------------------------
SpawnGroundWave = function(ClosestWaypoint, SpawnMultiplier, StartingGates) -- This generates the units that will be in the wave then spawns them

	local UnitID = '0';
	local UnitID2 = '0';
	local Quantity = 0;

	if ClosestWaypoint < 4 then
		UnitID = 'url0107';
		Quantity = (((ClosestWaypoint + 1) * 2) * SpawnMultiplier);
	elseif ClosestWaypoint <7 then
		UnitID = 'url0202';
		Quantity = ((ClosestWaypoint) * SpawnMultiplier);
	elseif ClosestWaypoint >= 7 then
		UnitID = 'url0303';
		Quantity = (((ClosestWaypoint)/2) * SpawnMultiplier);
	end

	if ClosestWaypoint < 4 then
		UnitID2 = 'url0103';
		#Quantity = (((ClosestWaypoint * 2) + 1) * SpawnMultiplier);
	elseif ClosestWaypoint <7 then
		UnitID2 = 'url0111';
		#Quantity = ((ClosestWaypoint) * SpawnMultiplier);
	elseif ClosestWaypoint < 9 then
		UnitID2 = 'url0304';
	elseif ClosestWaypoint >= 9 then
		UnitID2 = 'xel0306';
		#Quantity = (((ClosestWaypoint)/2) * SpawnMultiplier);
	end

	local ArmyID = "ARMY_SURVIVAL_ENEMY";
	local POS = ScenarioUtils.MarkerToPosition("GROUNDSPAWN_" .. ClosestWaypoint);
	local OrderID = 2;

	local GatesRemaining = GetGatesRemaining(ClosestWaypoint, 'GROUND');
	if GatesRemaining < StartingGates and StartingGates > 0 then
		Quantity = Quantity * (GatesRemaining/StartingGates);
	end

	Survival_SpawnUnit(UnitID, ArmyID, POS, OrderID, Quantity, UnitID2, 'GROUND')


end


-- air spawn
---------------------------------------------------------------------------------------------
SpawnAirWave = function(ClosestWaypoint, SpawnMultiplier) -- This generates the units that will be in the wave then spawns them

	local UnitID = '0';
	local UnitID2 = '0';
	local Quantity = 0;

	local ArmyID = "ARMY_SURVIVAL_ENEMY";
	local POS = ScenarioUtils.MarkerToPosition("AIRSPAWN_" .. ClosestWaypoint);
	local OrderID = 2;

	if ClosestWaypoint < 4 then
	UnitID = 'ura0103';
	#Quantity = ((ClosestWaypoint + 2) * SpawnMultiplier);
	elseif ClosestWaypoint <8 then
	UnitID = 'ura0203';
	#Quantity = (ClosestWaypoint * SpawnMultiplier);
	elseif ClosestWaypoint >= 8 then
	UnitID2 = 'ura0303';
	UnitID3 = 'xra0305';
	#Quantity = ((ClosestWaypoint)/2 * SpawnMultiplier);
	Survival_SpawnUnit(UnitID2, ArmyID, POS, OrderID, Quantity, UnitID3, 'AIR')
	end

	if ClosestWaypoint < 4 then
	UnitID2 = 'ura0102';
	Quantity = (((ClosestWaypoint + 1) * 2) * SpawnMultiplier);
	elseif ClosestWaypoint < 8 then
	UnitID2 = 'dra0202';
	Quantity = ((ClosestWaypoint) * SpawnMultiplier);
	elseif ClosestWaypoint >= 8 then
	UnitID = 'ura0304';
	UnitID2 = 'ura0303';
	Quantity = (((ClosestWaypoint + 1)/2) * SpawnMultiplier);
	end
	
	Survival_SpawnUnit(UnitID, ArmyID, POS, OrderID, Quantity, UnitID2, 'AIR')


end


-- naval spawn
---------------------------------------------------------------------------------------------
SpawnNavalWave = function(ClosestWaypoint, SpawnMultiplier, StartingGates) -- This generates the units that will be in the wave then spawns them

	local UnitID = '0';
	local UnitID2 = '0';
	local Quantity = 0;

	local navalRandom = math.random(10); 
	if navalRandom < 5 then	
	-- cut down on naval output, 'randomly'
	LOG("navalRandom is " .. navalRandom .. "spawning naval units, random < 5");	
		if ClosestWaypoint < 4 then
			UnitID = 'urs0103';
			#Quantity = (ClosestWaypoint * SpawnMultiplier);
		elseif ClosestWaypoint <6 then
			UnitID = 'urs0201';
			#Quantity = ((ClosestWaypoint + 1) * SpawnMultiplier);
		elseif ClosestWaypoint >= 6 then
			UnitID = 'urs0202';
			#Quantity = ((ClosestWaypoint/5) * SpawnMultiplier);
		end
		if ClosestWaypoint < 4 then
			UnitID2 = 'urs0203';
			Quantity = ((ClosestWaypoint / 2) * SpawnMultiplier);
		elseif ClosestWaypoint <8 then
			UnitID2 = 'xrs0204';
			Quantity = ((ClosestWaypoint / 4) * SpawnMultiplier);
		elseif ClosestWaypoint >= 8 then
			UnitID2 = 'urs0302';
			Quantity = (1 * SpawnMultiplier);
		end
		local ArmyID = "ARMY_SURVIVAL_ENEMY";
		local POS = ScenarioUtils.MarkerToPosition("NAVALSPAWN_" .. ClosestWaypoint);
		local OrderID = 2;
		local GatesRemaining = GetGatesRemaining(ClosestWaypoint, 'NAVAL');
		if GatesRemaining < StartingGates and StartingGates > 0 then
			Quantity = Quantity * (GatesRemaining/StartingGates);
		end
		Survival_SpawnUnit(UnitID, ArmyID, POS, OrderID, Quantity, UnitID2, 'NAVAL');
    end
	LOG("End of Naval spawn");
end


-- Gate Spawn Remaining count.  Checks on Quantam Gateway
-- check the use of AIUtils.GetOwnUnitsAroundPoint in /lua/ai/aiutilities.lua
------------------------------------------
GetGatesRemaining = function(WaypointNumber, Type)
	local spawnCategory = '';
	
	if Type == 'Naval' then
		spawnCategory = categories.urb0303
	elseif Type == 'GROUND' then
		spawnCategory = categories.urb0304
	elseif Type == 'AIR' then
		spawnCategory = categories.urb0302
	else
		spawnCategory = categories.urb0304  -- defualt
	end
    
	local units = AIUtils.GetOwnUnitsAroundPoint(GetArmyBrain("ARMY_SURVIVAL_ENEMY"), spawnCategory, ScenarioUtils.MarkerToPosition(Type .. "SPAWN_" .. WaypointNumber), 30 )
    return table.getn(units) 
end


-- Gate Spawn Remaining count.  Checks on Naval Factories
-- check the use of AIUtils.GetOwnUnitsAroundPoint in /lua/ai/aiutilities.lua
------------------------------------------
--GetNavalGatesRemaining = function(WaypointNumber)
--      local units = AIUtils.GetOwnUnitsAroundPoint(GetArmyBrain("ARMY_SURVIVAL_ENEMY"), categories.urb0303, ScenarioUtils.MarkerToPosition("NAVALSPAWN_" .. WaypointNumber), 30 )
--      return table.getn(units)
--end


-- Spawn units - 2 unit ID's passed.
---------------------------------------------------------------------------------------------
Survival_SpawnUnit = function(UnitID, ArmyID, POS, OrderID, Quantity, UnitID2, Type) -- blueprint, army, position, order

--	LOG("----- Survival MOD: SPAWNUNIT: Start function...")
	local PlatoonList = {}
	local Counter = 0

	while Counter < Quantity do
		local NewUnit = CreateUnitHPR(UnitID, ArmyID, POS[1], POS[2], POS[3], 0, Random(0,360),0)
		local NewUnit2 = CreateUnitHPR(UnitID2, ArmyID, POS[1], POS[2], POS[3], 0, Random(0,360),0)
		table.insert(PlatoonList, NewUnit) -- add unit to a platoon
		table.insert(PlatoonList, NewUnit2) -- add unit to a platoon
		Survival_PlatoonOrder(ArmyID, PlatoonList, OrderID, Type) -- give the unit orders
		Counter = Counter + 1
	end
end


-- Spawn units - spawn 1 unit
---------------------------------------------------------------------------------------------
Survival_SpawnUnit1 = function(UnitID, ArmyID, POS, OrderID, Quantity, Type) -- blueprint, army, position, order

--	LOG("----- Survival MOD: SPAWNUNIT: Start function...");
	local PlatoonList = {};
	local Counter = 0;

	while Counter < Quantity do
		local NewUnit = CreateUnitHPR(UnitID, ArmyID, POS[1], POS[2]+ (Random(0,2) * 10), POS[3], 0, Random(0,360),0);
		table.insert(PlatoonList, NewUnit); -- add unit to a platoon
		Survival_PlatoonOrder(ArmyID, PlatoonList, OrderID, Type); -- give the unit orders
		Counter = Counter + 1;
	end
end

-- Spawn units - spawn 1 unit
---------------------------------------------------------------------------------------------
Survival_SpawnBossUnit = function(UnitID, ArmyID, POS, OrderID, Quantity, Type) -- blueprint, army, position, order
	
	local aiBrain = GetArmyBrain(ArmyID);
	local aiPlatoon = aiBrain:MakePlatoon('','');
	local PlatoonList = {};
	local Counter = 0;
	
	while Counter < Quantity do
		local NewUnit = CreateUnitHPR(UnitID, ArmyID, POS[1], POS[2]+ (Random(0,2) * 10), POS[3], 0, Random(0,360),0);
		table.insert(PlatoonList, NewUnit); -- add unit to a platoon
		#Survival_PlatoonOrder(ArmyID, PlatoonList, OrderID, Type); -- give the unit orders
		aiBrain:AssignUnitsToPlatoon(aiPlatoon, PlatoonList, 'Attack', 'None') 
		aiPlatoon:AggressiveMoveToLocation(Survival_DefUnit:GetPosition())
		#aiPlatoon:AggressiveMovetoLocation(ScenarioUtils.MarkerToPosition("WAYPOINT_" .. TargetWaypoint ))
		Counter = Counter + 1;
	end
end


-- Surviva Platoon Order
--------------------------------------------------------------------------
Survival_PlatoonOrder = function(ArmyID, UnitList, OrderID, Type)

--	LOG("----- Survival MOD: PLATOON: Start function...");

	if (UnitList == nil) then
		return;
	end

	local aiBrain = GetArmyBrain(ArmyID); --"ARMY_SURVIVAL_ENEMY");
	local aiPlatoon = aiBrain:MakePlatoon('','');
	aiBrain:AssignUnitsToPlatoon(aiPlatoon, UnitList, 'Attack', 'None'); -- platoon, unit list, "mission" and formation

 	-- 1 center / 2 waypoint / 3 spawn
 	if (OrderID == 2) then -- attack move
		-- attack move to random path
		POS = Survival_DefUnit:GetPosition();
		aiPlatoon:AggressiveMoveToLocation(POS);
	elseif (OrderID == 1) then
		POS = Survival_DefUnit:GetPosition();
		aiPlatoon:MoveToLocation(POS, false);
 	#elseif (OrderID == 3) then -- patrol paths
	#	-- patrol to random path
	#	POS = Survival_GetPOS(2, 25);
	#	aiPlatoon:Patrol(POS);
	end
	if Type == 'AIR' then
		aiPlatoon:Patrol(Survival_DefUnit:GetPosition())
		aiPlatoon:Patrol(ScenarioUtils.MarkerToPosition("WAYPOINT_" .. TargetWaypoint))
	end
end


-- handles the emergency
---------------------------------------
EmergencyShield = function(Difficulty)

	local	EmergencyShieldCount = nil;

	if Difficulty == 1 then
		EmergencyShieldCount = 10;
	elseif Difficulty == 2 then 
		EmergencyShieldCount = 5;
	elseif Difficulty == 3 then 
		EmergencyShieldCount = 4;
	elseif Difficulty == 4 then 
		EmergencyShieldCount = 3;
	elseif Difficulty == 5 then 
		EmergencyShieldCount = 3;
	elseif Difficulty == 6 then 
		EmergencyShieldCount = 2;
	elseif Difficulty == 7 then 
		EmergencyShieldCount = 2;
	elseif Difficulty == 8 then 
		EmergencyShieldCount = 2;
	else 
		EmergencyShieldCount = 1;
	end

	LOG("Emergency Shield thread running with " .. EmergencyShieldCount .. " starting shields.");

		local bp ={
            ImpactEffects = 'SeraphimShieldHit01',
            ImpactMesh = '/effects/entities/ShieldSection01/ShieldSection01_mesh',
			Mesh = '/effects/entities/SeraphimShield01/SeraphimShield01_mesh',
            MeshZ = '/effects/entities/Shield01/Shield01z_mesh',
            RegenAssistMult = 60,
            ShieldEnergyDrainRechargeTime = 5,
            ShieldMaxHealth = 500000,
            ShieldRechargeTime = 160,
            ShieldRegenRate = 64,
            ShieldRegenStartTime = 1,
            ShieldSize = 15,
            ShieldVerticalOffset = -4,
	    }
	
	while not Survival_DefUnit:IsDead() do

		if Survival_DefUnit:GetHealth() < (Survival_DefUnit:GetMaxHealth() * .2) then
			BroadcastMSG("Spider Pig hull integrity CRITICAL!", 4);
			if EmergencyShieldCount > 0 then
				EmergencyShieldCount = EmergencyShieldCount - 1;
				BroadcastMSG("Emergency Sheild activated.  Spider Pig out.", 4);	
      			ForkThread(function() WaitTicks(1) Survival_DefUnit:CreateShield(bp) end);
				Survival_DefUnit:SetSpeedMult(0.01);
           		local RegenAmount = (((Survival_DefUnit:GetMaxHealth() * .2) - Survival_DefUnit:GetHealth())/15)
				local Buff = import('/lua/sim/Buff.lua')
				if not Buffs['EmergencyRegen'] then
				BuffBlueprint {
					Name = 'EmergencyRegen',
					DisplayName = 'EmergencyRegen',
					BuffType = 'ACUUPGRADEDMG',
					Stacks = 'ALWAYS',
					Duration = 30,
					Affects = {
						MaxHealth = {
							Add = 0,
							Mult = 1.0,
						},
						Regen = {
							Add = RegenAmount,
							Mult = 1.0,
						},
					 },
				} 	
				end
				Buff.ApplyBuff(Survival_DefUnit, 'EmergencyRegen')   
				LOG("Regen amount for emer shield is " .. RegenAmount)
				WaitSeconds(30)
				BroadcastMSG("Emergency Shield dissipated. Remaining Shield count: " .. EmergencyShieldCount, 4);
	   	        Survival_DefUnit:SetSpeedMult(0.5);
				Survival_DefUnit:DestroyShield();
			end
		if EmergencyShieldCount == 0 then
			Survival_DefUnit:SetSpeedMult(0.5); -- make sure speed is reset. 
			WaitSeconds(10)
		end
		end
		WaitSeconds(1)
	end
end


-- Controls the spawn of enemy XP units
---------------------------------------------------------------------
function BossBattles(BossSpawnMarkers)
	
	local NextBossBattle = 1;
	local WarningIssued = 'false';

	while not Survival_DefUnit:IsDead() and NextBossBattle <= TotalBossBattles do
		local Distance = VDist3(Survival_DefUnit:GetPosition(),ScenarioUtils.MarkerToPosition("BOSS_BATTLE_" .. NextBossBattle));	
			if (Distance < 50) and (WarningIssued == 'false') then
				if (NextBossBattle == 1 ) then
					BroadcastMSG("Dr Who has informed us that some Experimentals were built in this area. Spider Pig out.", 5);
					WarningIssued = 'true';
				elseif (NextBossBattle == 2) then
					BroadcastMSG("Intel pikcing up two unknown signatures.  HQ Out.", 5);
					WarningIssued = 'true';
				elseif (NextBossBattle == 3) then
					#BroadcastMSG("", 5); -- suprise
					WarningIssued = 'true';
				elseif (NextBossBattle == 4) then
					BroadcastMSG("Intel is picking up some strange readings. Proceed with caution. HQ Out.", 5);
					WarningIssued = 'true';
				end
			end
			if (Distance < 10) then				
				if (NextBossBattle == 1 ) then
					BroadcastMSG("Enemy Cybran Experimental sighted in the operational area.  Destroy it!", 5);
					local BSA = math.random(3);
					LOG("BSA is " .. BSA);
					if BSA == 1 then
						Survival_SpawnBossUnit('url0402', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_1"), 2, 1, 'urs0202', 'GROUND');
					elseif BSA == 2 then
						Survival_SpawnBossUnit('url0402', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_1A"), 2, 1, 'url0304', 'GROUND');						
					else 
						Survival_SpawnBossUnit('url0402', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_1B"), 2, 1, 'url0304', 'GROUND');
					end
					NextBossBattle = NextBossBattle + 1;
					WarningIssued = 'false';
				elseif (NextBossBattle == 2) then
					BroadcastMSG("Enemy Cybran Experimentals sighted in the operational area.  Destroy THEM!.", 5);
					Survival_SpawnBossUnit('url0402', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_" .. NextBossBattle), 2, 2, 'xrl0305', 'GROUND');
					NextBossBattle = NextBossBattle + 1;
				elseif (NextBossBattle == 3) then
					BroadcastMSG("A UEF traitor! What! Destroy that Fatboy!!", 5);
					Survival_SpawnBossUnit('uel0401', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_" .. NextBossBattle), 2, 2, 'xel0306', 'GROUND');
					NextBossBattle = NextBossBattle + 1;
				elseif (NextBossBattle == 4) then
					BroadcastMSG("Intel is still scrambled. We are only picking up TWO Experimentals. Something must be wrong.", 5);
					Survival_SpawnBossUnit('url0402', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_" .. NextBossBattle), 2, 3, 'url0304', 'GROUND');
					Survival_SpawnBossUnit('xrl0403', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_" .. NextBossBattle), 2, 1, 'xel0306', 'GROUND');
					--add in a bunch of air ;)
					local cWaypoint = GetClosestWaypoint();
					SpawnAirWave((cWaypoint - 3), 1);
					SpawnAirWave((cWaypoint - 2), 1);					
					SpawnAirWave((cWaypoint - 1), 1);
					--end air
					NextBossBattle = NextBossBattle + 1;
				elseif (NextBossBattle == 5) then	
					Survival_SpawnBossUnit('ura0401', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_5" .. NextBossBattle), 2, 1, 'xel0306', 'AIR');
					Survival_SpawnBossUnit('ura0401', 'ARMY_SURVIVAL_ENEMY', ScenarioUtils.MarkerToPosition("BOSSSPAWN_5A" .. NextBossBattle), 2, 1, 'xra0305', 'AIR');
					NextBossBattle = NextBossBattle + 1;
				end
			end
		WaitSeconds(10)
	end
end


-- GetMarker function , had duplicates, deleted second
------------
function GetMarker(MarkerName)
	return Scenario.MasterChain._MASTERCHAIN_.Markers[MarkerName];
end

-- Controls map expansion
---------------------------------------------------------------------
function ExpansionCheck()
	LOG("In Expansion Check: CurrentArea: " .. CurrentArea .. "  TotalMapExpansions" .. TotalMapExpansions);
	--local NextMapExpansion = 1;

	while not Survival_DefUnit:IsDead() and CurrentArea <= TotalMapExpansions  do
		local Distance = VDist3(Survival_DefUnit:GetPosition(),ScenarioUtils.MarkerToPosition("MAP_EXPANSION_" .. CurrentArea));
			
			if (Distance < 10) then
				if CurrentArea == 1 then
					ForkThread(SpeedBoost)
				end
				CurrentArea = CurrentArea + 1;
				BroadcastMSG("Waypoint reached.  Expanding the operational area.  HQ out.", 5);

				--SpawnDefenses(CurrentArea);
				if (CurrentArea == 3) then
					ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ENEMY','GROUP_22');--second expansion
					LOG("Spawning extra units");
				end

 				ScenarioFramework.SetPlayableArea('AREA_' .. (CurrentArea), false);
				 
				
				--ForkThread(Waypoint_Tracker);
			end
		LOG("Next map expansion " .. CurrentArea .. " at distance " .. Distance);
		WaitSeconds(10);
	end
end


-- spawns a defense object
-- Monkeylord - need to figure out proper HP.
--------------------------------------------------------------------------
Survival_SpawnDef = function()

	LOG("----- Survival MOD: Initializing defense object...");

	local POS = ScenarioUtils.MarkerToPosition("WAYPOINT_0");

	Survival_DefUnit = CreateUnitHPR('url0402', "ARMY_SURVIVAL_ALLY", POS[1], POS[2], POS[3], 0,180,0);  
	ForkThread(Waypoint_Tracker)
	ForkThread(VictoryCheck)
	
	Survival_DefUnit:SetReclaimable(false);
	Survival_DefUnit:SetCapturable(false);
	Survival_DefUnit:SetSpeedMult(0.5);

	Survival_DefUnit:HideBone('Center_Turret', true)
	Survival_DefUnit:HideBone('Center_Turret_Muzzle', true)
	Survival_DefUnit:SetWeaponEnabledByLabel('MainGun', false)

	#Survival_DefUnit:GetWeaponByLabel('RightAntiAirMissile'):ChangeRateOfFire(2) -- if I want to change rof
	Survival_DefUnit:GetWeaponByLabel('RightLaserTurret'):AddDamageMod(200) -- if I want to change dmg
	Survival_DefUnit:GetWeaponByLabel('LeftLaserTurret'):AddDamageMod(200)
	Survival_DefUnit:GetWeaponByLabel('RightAntiAirMissile'):AddDamageMod(100)
	Survival_DefUnit:GetWeaponByLabel('LeftAntiAirMissile'):AddDamageMod(100)
	
	Survival_DefUnit:SetCustomName("Spider Pig");

	local Buff = import('/lua/sim/Buff.lua')
	
        if not Buffs['SeraphimACUDamageStabilizationMoo'] then
            BuffBlueprint {
                Name = 'SeraphimACUDamageStabilizationMoo',
                DisplayName = 'SeraphimACUDamageStabilizationMoo',
                BuffType = 'ACUUPGRADEDMG',
                Stacks = 'ALWAYS',
                Duration = -1,
                Affects = {
                    MaxHealth = {
                        Add = 10000,-- 45000 base  - adding 10k 
                        Mult = 1.0,
                    },
                    Regen = {
                        Add = 90,
                        Mult = 1.0,
                    },
                },
            } 	
		end
        Buff.ApplyBuff(Survival_DefUnit, 'SeraphimACUDamageStabilizationMoo')    
	
	Survival_DefUnit:SetIntelRadius('Vision', 90);
	Survival_DefUnit:SetIntelRadius('Radar', 200);

	-- when the def object dies
	Survival_DefUnit.OldOnKilled = Survival_DefUnit.OnKilled;
	 
	Survival_DefUnit.OnKilled = function(self, instigator, type, overkillRatio)
		BroadcastMSG("The Doctor has been lost.  Our attempts failed to save him have failed.", 8);
		self.OldOnKilled(self, instigator, type, overkillRatio);
		Survival_GameState = 3;
		for i, army in ListArmies() do
			if (army == "ARMY_1" or army == "ARMY_2" or army == "ARMY_3" or army == "ARMY_4" or army == "ARMY_5" or army == "ARMY_6" or army == "ARMY_7" or army == "ARMY_8") then
				--killAllCommanders(army);
				GetArmyBrain(army):OnDefeat();
			end
		end
	end
	LOG("Spawn Defensive unit complete");
end


-- Spawn random enemy and or firebases on map start up
-- adds some varierty, falvor to the map, to change experience from time to time.  Nothing major, but meant to keep you on yoru toes a little more.
Survival_SpawnEnemyRandoms = function()

	LOG("In Survival_SpawnEnemyRandoms");
	local RandomQuantity = math.random(10);
	LOG("RandomQuantity is " .. RandomQuantity);
	
	-- change start of map randomly.  Want to add third option.
	if RandomQuantity < 6 then
		ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ENEMY','GROUP_1'); --little AA and torp in starting bay
	else 
		ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ENEMY','GROUP_3'); -- 3rd set of PDs by WP1.
	end
	
	-- spawn either the artillary up on top left, or a set of PDs.  Which will it be???
	if RandomQuantity < 6 then
		ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ENEMY','GROUP_2'); --AA and Artillary
	else
		ScenarioUtils.CreateArmyGroup('ARMY_SURVIVAL_ENEMY','GROUP_4'); --PD setup	
	end
end

--makes the survival unit go faster if it has full hp
-----------------------
SpeedBoost = function()
	local CurrentSpeed = .5
	while not Survival_DefUnit:IsDead() do
		local MaxHealth = Survival_DefUnit:GetMaxHealth()
		local Health1 = Survival_DefUnit:GetHealth()
		if Health1 < (MaxHealth * .8) and CurrentSpeed > .5 then
			Survival_DefUnit:SetSpeedMult(.4)
			CurrentSpeed = .5
			BroadcastMSG("Spider Pig is taking damage.  Reverting power back to shields.", 4)
		elseif Health1 >= (MaxHealth * .9) and CurrentSpeed == .5 then
			WaitSeconds(10)
			Health1 = Survival_DefUnit:GetHealth()
			if Health1 >= (MaxHealth * .9) and CurrentSpeed > .5 then
				Survival_DefUnit:SetSpeedMult(.5)
				CurrentSpeed = .5
				BroadcastMSG("Spider Pig is taking damage. Reverting power back to shields.", 4)
			elseif Health1 == MaxHealth then
			WaitSeconds(10)
			Health1 = Survival_DefUnit:GetHealth()
				if Health1 >= (MaxHealth * .9)  and CurrentSpeed > .5  then
					Survival_DefUnit:SetSpeedMult(.5)
					BroadcastMSG("Spider Pig is taking damage. Reverting power back to shields.", 4)
				elseif Health1 == MaxHealth then
					Survival_DefUnit:SetSpeedMult(.5)
					CurrentSpeed = .5
					BroadcastMSG("Spider Pig is in good shape.  But NOT Hitting the gas.", 4)
				end
			end
		end
		WaitSeconds(10)
	end
end
